import os
import sys
from collections import defaultdict
from enum import Enum
from functools import lru_cache
from typing import Type, List, cast, Optional, Dict

import spacy
from pydantic import BaseModel, Field
from pymultirole_plugins.util import comma_separated_to_list
from pymultirole_plugins.v1.annotator import AnnotatorParameters, AnnotatorBase
from pymultirole_plugins.v1.schema import Document, Span, Annotation, Term
from scispacy.linking_utils import Entity
from spacy.cli.download import download_model, get_compatibility, get_version
from spacy.language import Language
from spacy.util import run_command
from wasabi import msg

_home = os.path.expanduser('~')
xdg_cache_home = os.environ.get('XDG_CACHE_HOME') or os.path.join(_home, '.cache')
cache_dir = os.path.join(xdg_cache_home, 'scispacy')
os.environ["SCISPACY_CACHE"] = cache_dir


class SpacyModelType(str, Enum):
    base = 'base'
    sci = 'sci'


class SpacyModelSize(str, Enum):
    sm = 'sm'
    md = 'md'
    lg = 'lg'


class SciSpacyLinker(str, Enum):
    umls = 'umls'
    mesh = 'mesh'


class SpacyLabel(str, Enum):
    # fr,de,pt,it
    PER = "PER"
    MISC = "MISC"
    # en,zh,nl
    LOC = "LOC"
    ORG = "ORG"
    PERSON = "PERSON"
    CARDINAL = "CARDINAL"
    DATE = "DATE"
    EVENT = "EVENT"
    FAC = "FAC"
    GPE = "GPE"
    LANGUAGE = "LANGUAGE"
    LAW = "LAW"
    MONEY = "MONEY"
    NORP = "NORP"
    ORDINAL = "ORDINAL"
    PERCENT = "PERCENT"
    PRODUCT = "PRODUCT"
    QUANTITY = "QUANTITY"
    TIME = "TIME"
    WORK_OF_ART = "WORK_OF_ART"
    # scispacy
    ENTITY = "ENTITY"


class SpacyNERParameters(AnnotatorParameters):
    model_type: SpacyModelType = Field(SpacyModelType.base,
                                       description="""Type of NER spacy model to use for NER extraction, can be:
                       <li>`base` for basic entities like location/person/organization provided by [spacy](https://spacy.io/models) or
                       <li>`sci` for biomedical entities provided by [scispacy](https://allenai.github.io/scispacy)""")
    mapping: Dict[str, List[SpacyLabel]] = Field(
        None,
        description="Map a label to a list of spacy NER labels",
        extra="key:label",
    )
    model_size: SpacyModelSize = Field(SpacyModelSize.sm,
                                       description="""Size of NER spacy model to use for NER extraction, can be:
                       <li>`sm` for small,
                       <li>`md` for medidum or
                       <li>`lg` for large""",
                                       extra="advanced,internal")
    scispacy_linker: Optional[SciSpacyLinker] = Field(None,
                                                      description="""Name of the [scispacy linker](https://github.com/allenai/scispacy) which performs linking to a knowledge base.<br/>
                                                      <b>Only valid with scispacy models</b>""",
                                                      extra="advanced")
    types_filter: Optional[str] = Field("", description="""Comma-separated list of UMLS [semantic types](https://lhncbc.nlm.nih.gov/ii/tools/MetaMap/documentation/SemanticTypesAndGroups.html) to keep.<br/>
                                                      For example `T028,T116,T123`<br/>
                                                      <b>Only valid with scispacy models</b>""",
                                        extra="advanced")
    mention_threshold: Optional[float] = Field(0.85,
                                               description="""The threshold that a entity candidate must reach to be added.<br/>
                                                      <b>Only valid with scispacy models</b>""",
                                               extra="advanced")
    mention_max: Optional[int] = Field(1,
                                       description="""The maximum number of entities which will be returned for a given mention.<br/>
                                                      <b>Only valid with scispacy models</b>""",
                                       extra="advanced")


SUPPORTED_LANGUAGES = "en,fr,de,nl,es,pt,it,zh,ru,ja"


class SpacyNERAnnotator(AnnotatorBase):
    __doc__ = """[Spacy](https://spacy.io/api/entityrecognizer) annotator.
    #need-segments
    #languages:""" + SUPPORTED_LANGUAGES

    def annotate(self, documents: List[Document], parameters: AnnotatorParameters) \
            -> List[Document]:
        params: SpacyNERParameters = \
            cast(SpacyNERParameters, parameters)
        supported_languages = comma_separated_to_list(SUPPORTED_LANGUAGES)
        if params.mapping is not None:
            spacy2labels = defaultdict(list)
            for k, label_lst in params.mapping.items():
                for label in label_lst:
                    spacy2labels[label].append(k)
        else:
            spacy2labels = None
        for document in documents:
            lang = document_language(document, None)
            if lang is None or lang not in supported_languages:
                raise AttributeError(f"Metadata language {lang} is required and must be in {SUPPORTED_LANGUAGES}")
            # Create parsing context with time and language information
            nlp, linker = get_nlp(lang, params.model_type, params.model_size, params.scispacy_linker)
            types_filter = comma_separated_to_list(params.types_filter.strip())
            types_as_set = set(types_filter or [])
            document.annotations = []
            if not document.sentences:
                document.sentences = [Span(start=0, end=len(document.text))]
            sents = [document.text[s.start:s.end] for s in document.sentences]
            asents = nlp.pipe(sents)
            for sent, asent in zip(document.sentences, asents):
                for ent in asent.ents:
                    if spacy2labels is None or ent.label_ in spacy2labels:
                        start = sent.start + ent.start_char
                        end = sent.start + ent.end_char
                        terms = []
                        if linker is not None:
                            for umls_ent in ent._.kb_ents:
                                score = umls_ent[1]
                                if score >= params.mention_threshold:
                                    kbent: Entity = linker.kb.cui_to_entity[umls_ent[0]]
                                    types = types_as_set.intersection(kbent.types)
                                    if types or len(types_as_set) == 0:
                                        props = {'types': kbent.types if len(types_as_set) == 0 else list(types),
                                                 'aliases': kbent.aliases}
                                        term = Term(identifier=umls_ent[0], lexicon=params.scispacy_linker.value,
                                                    preferredForm=kbent.canonical_name, score=score, properties=props)
                                        if len(terms) < params.mention_max:
                                            terms.append(term)
                                        else:
                                            break
                        if spacy2labels is None:  # same as before
                            document.annotations.append(Annotation(start=start, end=end,
                                                                   labelName=ent.label_.lower(),
                                                                   label=ent.label_,
                                                                   text=document.text[start:end],
                                                                   terms=terms if linker is not None else None))
                        else:
                            for lname in spacy2labels[ent.label_]:
                                document.annotations.append(Annotation(start=start, end=end,
                                                                       labelName=lname,
                                                                       text=document.text[start:end],
                                                                       terms=terms if linker is not None else None))
        return documents

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return SpacyNERParameters


def document_language(doc: Document, default: str = None):
    if doc.metadata is not None and 'language' in doc.metadata:
        return doc.metadata['language']
    return default


@lru_cache(maxsize=None)
def get_nlp(lang, model_type, model_size, linker=None):
    model = lang
    if model_type == SpacyModelType.sci:
        if lang != 'en':
            raise AttributeError("scispacy model is only available in english")
        else:
            model += f"_core_{model_type.value}"
    else:
        model += "_core"
        if lang in ['en', 'zh']:
            model += "_web"
        else:
            model += "_news"
    model += f"_{model_size.value}"
    nlp: Language = load_spacy_model(model)
    if model_type == SpacyModelType.sci and linker is not None:
        from scispacy.linking import EntityLinker
        linker: EntityLinker = nlp.add_pipe("scispacy_linker", config={"linker_name": linker.value})
    return nlp, linker


def load_spacy_model(model, *pip_args):
    if '_sci_' in model:
        model_name = model
        download_sci_model(model)
    else:
        suffix = "-py3-none-any.whl"
        dl_tpl = "{m}-{v}/{m}-{v}{s}#egg={m}=={v}"
        model_name = model
        compatibility = get_compatibility()
        version = get_version(model_name, compatibility)
        download_model(dl_tpl.format(m=model_name, v=version, s=suffix), pip_args)
    msg.good(
        "Download and installation successful",
        f"You can now load the package via spacy.load('{model_name}')",
    )
    # If a model is downloaded and then loaded within the same process, our
    # is_package check currently fails, because pkg_resources.working_set
    # is not refreshed automatically (see #3923). We're trying to work
    # around this here be requiring the package explicitly.
    require_package(model_name)
    return spacy.load(model_name)


def require_package(name):
    try:
        import pkg_resources

        pkg_resources.working_set.require(name)
        return True
    except:  # noqa: E722
        return False


def download_sci_model(
        filename: str, user_pip_args=None) -> None:
    download_url = f"https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.4.0/{filename}-0.4.0.tar.gz"
    pip_args = list(user_pip_args) if user_pip_args is not None else []
    cmd = [sys.executable, "-m", "pip", "install"] + pip_args + [download_url]
    run_command(cmd)



import json
import string
from collections import defaultdict, Counter
from pathlib import Path
from typing import List

import pandas as pd
import pytest as pytest
from fuzzywuzzy import fuzz
from fuzzywuzzy import process
from pybel.parser import BELParser
from pymultirole_plugins.v1.schema import Document, DocumentList, Sentence

from pyannotators_spacyner.spacyner import SpacyNERAnnotator, SpacyNERParameters


def test_spacyner():
    model = SpacyNERAnnotator.get_model()
    model_class = model.construct().__class__
    assert model_class == SpacyNERParameters
    annotator = SpacyNERAnnotator()
    parameters = SpacyNERParameters()
    docs: List[Document] = annotator.annotate([Document(
        text="Paris is the capital of France and Emmanuel Macron is the president of the French Republic.",
        metadata={'language': 'en'})], parameters)
    doc0 = docs[0]
    assert len(doc0.annotations) == 4
    paris = doc0.annotations[0]
    france = doc0.annotations[1]
    macron = doc0.annotations[2]
    republic = doc0.annotations[3]
    assert paris.label == 'GPE'
    assert france.label == 'GPE'
    assert macron.label == 'PERSON'
    assert republic.label == 'GPE'

    parameters.mapping = {
        "location": ["LOC", "GPE"],
        "person": ["PERSON"],
        "organization": ["ORG"]
    }
    docs: List[Document] = annotator.annotate(
        [
            Document(
                text="Paris is the capital of France and Emmanuel Macron is the president of the French Republic.",
                metadata={'language': 'en'},
            )
        ],
        parameters
    )
    doc0 = docs[0]
    assert len(doc0.annotations) == 4
    paris = doc0.annotations[0]
    france = doc0.annotations[1]
    macron = doc0.annotations[2]
    republic = doc0.annotations[3]
    assert paris.labelName == "location"
    assert france.labelName == "location"
    assert macron.labelName == "person"
    assert republic.labelName == "location"


def test_scispacyner():
    model = SpacyNERAnnotator.get_model()
    model_class = model.construct().__class__
    assert model_class == SpacyNERParameters
    annotator = SpacyNERAnnotator()
    parameters = SpacyNERParameters(model_type='sci')
    docs: List[Document] = annotator.annotate([Document(
        text="Myeloid derived suppressor cells (MDSC) are immature myeloid cells with immunosuppressive activity.\nThey accumulate in tumor-bearing mice and humans with different types of cancer, including hepatocellular carcinoma (HCC).",
        metadata={'language': 'en'})], parameters)
    doc0 = docs[0]
    assert len(doc0.annotations) == 12


def to_camel_case(s):
    return s[0].lower() + string.capwords(s, sep='_').replace('_', '')[1:] if s else s


@pytest.mark.skip(reason="Not a test")
def test_scai_parse():
    testdir = Path(__file__).parent / 'data'
    excel_file = testdir / "all_bel_relations.xlsx"
    bel_df = pd.read_excel(excel_file).fillna(value="")
    docs = {}
    docids = Counter()
    for index, row in bel_df.iterrows():
        subject = row['Subject']
        relation = row['Relation']
        object = row['Object']
        docid = str(row['PubMedID'])
        sentence = row['EvidenceSentence']
        title = row['PublicationTitle']
        sent = subject + " " + to_camel_case(relation) + " " + object
        docids.update([docid])
        docinc = docids[docid]
        doc = docs.get((docid, sentence))
        if doc is None:
            doc = Document(identifier=f"{docid}-{docinc}", title=title, text=sentence,
                           sentences=[Sentence(start=0, end=len(sentence))], metadata={"bel_sentences": [sent]})
            docs[(docid, sentence)] = doc
        else:
            doc.metadata['bel_sentences'].append(sent)
        # parsed_sent = pybel.parse(sent)
    dl = DocumentList(__root__=list(docs.values()))
    json_file = testdir / "all_bell_relations.json"
    with json_file.open("w") as fout:
        print(dl.json(exclude_none=True, exclude_unset=True, indent=2), file=fout)


@pytest.mark.skip(reason="Not a test")
def test_scai():
    annotator = SpacyNERAnnotator()
    parameters = SpacyNERParameters(model_type='sci', scispacy_linker='umls',
                                    types_filter=','.join(
                                        ['T028', 'T116', 'T123', 'T126', 'T043', 'T047', 'T121', 'T109',
                                         'T026', 'T025', 'T129', 'T192', 'T044', 'T048', 'T043', 'T131',
                                         'T125', 'T043', 'T130', 'T196', 'T005', 'T008', 'T010', 'T012',
                                         'T013', 'T014', 'T015', 'T016', 'T017', 'T018', 'T021', 'T022',
                                         'T023', 'T024', 'T025', 'T026'
                                         ]))
    parameters = SpacyNERParameters(model='en_core_sci_sm', scispacy_linker='umls')
    testdir = Path(__file__).parent / 'data'
    json_file = testdir / "all_bell_relations.json"
    with json_file.open("r") as fin:
        docs = json.load(fin)
    docs = [Document(**doc) for doc in docs]
    docs = annotator.annotate(docs, parameters)
    json_file = testdir / f"all_bell_relations_annotated_{parameters.scispacy_linker.value}.json"
    dl = DocumentList(__root__=docs)
    with json_file.open("w") as fout:
        print(dl.json(exclude_none=True, exclude_unset=True, indent=2), file=fout)


def find_in_forms(el, forms):
    for k in el:
        if 'name' in el[k]:
            name = el[k]['name']
            if name.lower() in forms:
                return k
    return None


def fuzzy_find_in_forms(el, forms, threshold=85):
    for k in el:
        if isinstance(el[k], dict):
            if 'name' in el[k]:
                name = el[k]['name']
                extracted = process.extractOne(name.lower(), forms, scorer=fuzz.WRatio)
                if extracted[1] > threshold:
                    copy_el = {k1: v1 for k1, v1 in el.items() if k1 in ['function', k]}
                    return copy_el
        elif k == 'members' and isinstance(el[k], list):
            for subel in el[k]:
                ret = fuzzy_find_in_forms(subel, forms, threshold)
                if ret is not None:
                    return ret
    return None


def add2rows(el, rows):
    row = {}
    for k in el:
        if k == "function":
            row[k] = el[k]
        elif 'namespace' in el[k]:
            row[k] = el[k]['namespace']
    rows.append(row)


greek_alphabet = 'ΑαΒβΓγΔδΕεΖζΗηΘθΙιΚκΛλΜμΝνΞξΟοΠπΡρΣσςΤτΥυΦφΧχΨψΩω'
latin_alphabet = 'AaBbGgDdEeZzHhJjIiKkLlMmNnXxOoPpRrSssTtUuFfQqYyWw'
greek2latin = str.maketrans(greek_alphabet, latin_alphabet)


@pytest.mark.skip(reason="Not a test")
def test_scai_eval():
    testdir = Path(__file__).parent / 'data'
    linker = "umls"
    bel_parser = BELParser(skip_validation=True, citation_clearing=False, allow_naked_names=True)
    log_file = testdir / f"all_bell_relations_annotated_{linker}.log"
    with log_file.open("w") as log:
        json_file = testdir / f"all_bell_relations_annotated_{linker}.json"
        with json_file.open("r") as fin:
            docs = json.load(fin)
        docs = [Document(**doc) for doc in docs]
        rows = []
        functions = defaultdict(list)
        ok = 0
        for doc in docs:
            bel_sents = doc.metadata.get("bel_sentences", [])
            bels = []
            for bel_sent in bel_sents:
                try:
                    bels.append(bel_parser.parse(bel_sent))
                except BaseException:
                    print(f"Invalid BEL sentence: {bel_sent}")
                    bels.append(None)
            doc.metadata["bels"] = bels
            for a in doc.annotations:
                a.properties = defaultdict(list)
                forms = set()
                forms.add(a.text.lower())
                forms.add(a.text.translate(greek2latin).lower())
                for term in a.terms:
                    forms.add(term.preferredForm.lower())
                    if term.properties:
                        forms.update([f.lower() for f in term.properties.get('aliases', [])])
                a.properties['forms'] = list(forms)

            new_bels = []
            new_bel_sentences = []
            for ibel, bel in enumerate(doc.metadata['bels']):
                if bel is not None:
                    bel_sentence = doc.metadata['bel_sentences'][ibel]
                    src_found, tgt_found = False, False
                    source = bel['source']
                    add2rows(source, rows)
                    target = bel['target']
                    add2rows(target, rows)
                    for a in doc.annotations:
                        forms = a.properties['forms']
                        copy_source = fuzzy_find_in_forms(source, forms)
                        if copy_source is not None:
                            a.properties['source'].append(json.dumps(copy_source))
                            a.label = copy_source['function']
                            a.labelName = a.label.lower()
                            if a.terms:
                                for term in a.terms:
                                    functions[source['function']].extend(term.properties['types'])
                            if 'concept' in copy_source:
                                src_found = True
                        copy_target = fuzzy_find_in_forms(target, forms)
                        if copy_target is not None:
                            a.properties['target'].append(json.dumps(copy_target))
                            a.label = copy_target['function']
                            a.labelName = a.label.lower()
                            if a.terms:
                                for term in a.terms:
                                    functions[source['function']].extend(term.properties['types'])
                            if 'concept' in copy_target:
                                tgt_found = True
                    if src_found and tgt_found:
                        new_bels.append(bel)
                        new_bel_sentences.append((bel_sentence))
                    else:
                        print(f"Not found in sentence: {doc.text}", file=log)
                        print(f"BEL sentence: {bel_sentence}", file=log)
                        if not src_found:
                            print(f"\t- {source}", file=log)
                        if not tgt_found:
                            print(f"\t- {target}", file=log)
            del doc.metadata['bels']
            doc.metadata['bel_sentences'] = new_bel_sentences
            if len(new_bels) > 0:
                ok += 1
                doc.metadata['corpus'] = "train"
            else:
                doc.metadata['corpus'] = "test"
            for a in doc.annotations:
                del a.properties['forms']
                if 'source' in a.properties:
                    a.properties['source'] = [json.loads(x) for x in list(set(a.properties['source']))]
                if 'target' in a.properties:
                    a.properties['target'] = [json.loads(x) for x in list(set(a.properties['target']))]
        print(f"{ok}/{len(docs)} complete sentences")
        c = Counter()
        for function, flist in functions.items():
            print(function)
            print(Counter(flist))
            c = c + Counter(flist)
        print(c, file=log)
        print([k for k, v in c.items() if v >= 100])
        row_file = testdir / f"all_bell_relations_matched_{linker}.xlsx"
        df = pd.DataFrame.from_records(rows)
        df['function'].value_counts()
        df['concept'].value_counts()
        df['location'].value_counts()
        df.groupby('function').count()
        df.to_excel(row_file)
        json_file = testdir / f"all_bell_relations_matched_{linker}.json"
        dl = DocumentList(__root__=docs)
    with json_file.open("w") as fout:
        print(dl.json(exclude_none=True, exclude_unset=True, indent=2), file=fout)
